/*
Copyright (c) 2014, Intel Corporation
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
3. Neither the name of the Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <mathimf.h>
#include <omp.h>
#include <stdio.h>

#define NUM_STEPS 2048
#define OPT_N 5000*61*60
const float      RISKFREE = 0.06f;
const float    VOLATILITY = 0.10f;


__forceinline
float Exp( float x)
{
   return expf( x);
}

__forceinline
double Exp( double x)
{
   return exp( x);
}

__forceinline
float Sqrt( float x)
{
   return sqrtf( x);
}

__forceinline
double Sqrt( double x)
{
   return sqrt( x);
}



template <class Basetype>
void BinomialTemplate(
	Basetype *CallResult,
	Basetype *S,
	Basetype *X,
	Basetype *T)
{
//	kmpc_set_defaults("KMP_AFFINITY=granularity=fine,proclist=[1-240:1],explicit");
#pragma omp parallel for
	for(int opt = 0; opt < OPT_N; opt++)
	{
		__attribute__((align(4096)))Basetype Call[NUM_STEPS + 1];
		const Basetype       Sx = S[opt];
		const Basetype       Xx = X[opt];
		const Basetype       Tx = T[opt];

		const Basetype      dt = Tx / static_cast<Basetype>(NUM_STEPS);

		const Basetype     vDt = VOLATILITY * Sqrt(dt);
		const Basetype     rDt = RISKFREE * dt;

		const Basetype      If = Exp(rDt);
		const Basetype      Df = Exp(-rDt);

		const Basetype       u = Exp(vDt);
		const Basetype       d = Exp(-vDt);
		const Basetype      pu = (If - d) / (u - d);
		const Basetype      pd = 1.0f - pu;
		const Basetype  puByDf = pu * Df;
		const Basetype  pdByDf = pd * Df;
#pragma simd
		for(int i = 0; i <= NUM_STEPS; i++)
		{
			Basetype d = Sx * expf(vDt * (2.0 * i - NUM_STEPS)) - Xx;
			Call[i] =  (d > 0) ? d : 0;
		}
		for(int i = NUM_STEPS; i > 0; i--)
#pragma simd
#pragma unroll(4)
                   for(int j = 0; j <= i - 1; j++)
			Call[j] = puByDf * Call[j + 1] + pdByDf * Call[j];
		CallResult[opt] = (Basetype)Call[0];
	}

}
