/*
Copyright (c) 2014, Intel Corporation
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
3. Neither the name of the Intel Corporation nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mathimf.h>
#include <omp.h>
#include <sys/time.h>
#include <time.h>
#include "dvec.h"
#pragma offload_attribute(push,target(mic))
#include "BinomialGeneric.h"
#pragma offload_attribute(pop)
//#include <tbb/scalable_allocator.h>
#define scalable_aligned_malloc _mm_malloc
#define scalable_aligned_free _mm_free
#include <iostream>

#define SIMDALIGN 4096
char signal_var;
inline float RandFloat(float low, float high){
    float t = (float)rand() / (float)RAND_MAX;
    return (1.0f - t) * low + t * high;
}

inline double RandDouble(double low, double high){
    double t = (double)rand() / (double)RAND_MAX;
    return (1.0 - t) * low + t * high;
}

double second()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (double)tv.tv_sec + (double)tv.tv_usec / 1000000.0;
}

union FPARRAY 
{
	float  *SPData;
	double *DPData;
}; 

#define RANDSEED		123
template <class Basetype>
void BlackScholesTemplate(
    double *CallResult,
    Basetype *StockPrice,
    Basetype *OptionStrike,
    Basetype *OptionYears,
    Basetype Riskfree,
    Basetype Volatility,
    int optN
)
{
    for(int opt = 0; opt < optN; opt++)
        BlackScholesBody(
		CallResult[opt],
		StockPrice[opt],
		OptionStrike[opt],
		OptionYears[opt],
		Riskfree,
		Volatility
	);
}

int main(int argc, char* argv[])
{
   double 
        *CallResultMaster;
   FPARRAY 
        CallResult,
        StockPrice,
        OptionStrike,
        OptionYears;
   double
        delta, ref, sum_delta, sum_ref, max_delta, L1norm, sumReserve, errorVal;
   int 
	   i, mem_size, rand_size, verbose=1, doubleprecision=0; 
	double 
		sTime, eTime;
  
	const int TIMESTEP_N = 2048;

	if (argc > 3) {
        	printf("usage: Binomial <verbose> <precision>  where verbose = validation formation, precision = 1 for double, 0 for single. \n");
        	exit(1);
    }
	if (argc == 1) {
		verbose = 0;
		doubleprecision = 0;
	}
	if (argc == 2) {
		verbose = atoi(argv[1]);
		doubleprecision = 0;
	}
	if (argc == 3) {
		verbose = atoi(argv[1]);
		doubleprecision =  atoi(argv[2]);
	}

	if (doubleprecision)
		mem_size = sizeof(double)*OPT_N*2;
	else
		mem_size = sizeof(float)*OPT_N*2;

	if (verbose)
	{
		printf("Initializing data...\n");
		printf("...allocating memory for options.\n");
		// Master Data are always in double precision
		CallResultMaster = (double *)scalable_aligned_malloc(sizeof(double)*OPT_N*2, SIMDALIGN);
	}

	if (doubleprecision)
	{
		CallResult.DPData	= (double *)scalable_aligned_malloc(mem_size, SIMDALIGN);
		StockPrice.DPData	= (double *)scalable_aligned_malloc(mem_size, SIMDALIGN);
		OptionStrike.DPData	= (double *)scalable_aligned_malloc(mem_size, SIMDALIGN);
		OptionYears.DPData	= (double *)scalable_aligned_malloc(mem_size, SIMDALIGN);
	}
	else 
	{
		CallResult.SPData	= (float *)scalable_aligned_malloc(mem_size, SIMDALIGN);
		StockPrice.SPData	= (float *)scalable_aligned_malloc(mem_size, SIMDALIGN);
		OptionStrike.SPData	= (float *)scalable_aligned_malloc(mem_size, SIMDALIGN);
		OptionYears.SPData	= (float *)scalable_aligned_malloc(mem_size, SIMDALIGN);
	}
	if (verbose)
	{
		printf("...generating input data in mem.\n");
		srand(5347);
	}
	for(i = 0; i < OPT_N*2; i++)
	{
		if (doubleprecision)
		{
			CallResult.DPData[i]    = 0.0;
			StockPrice.DPData[i]    = RandDouble(5.0, 50.0);
			OptionStrike.DPData[i]  = RandDouble(10.0, 25.0);
			OptionYears.DPData[i]   = RandDouble(1.0, 5.0);
		}
		else 
		{
			CallResult.SPData[i]    = 0.0;
			StockPrice.SPData[i]    = RandFloat(5.0f, 50.0f);
			OptionStrike.SPData[i]  = RandFloat(10.0f, 25.0f);
			OptionYears.SPData[i]   = RandFloat(1.0f, 5.0f);
		}
	}

	if (verbose)
	{
		printf("Creating the reference result...\n");
		if (doubleprecision)
		{
			BlackScholesTemplate<double>(
				CallResultMaster,
				StockPrice.DPData,
				OptionStrike.DPData,
				OptionYears.DPData,
				RISKFREE,
				VOLATILITY,
				OPT_N*2
			);
		}
		else 
		{
			BlackScholesTemplate<float>(
				CallResultMaster,
				StockPrice.SPData,
				OptionStrike.SPData,
				OptionYears.SPData,
				RISKFREE,
				VOLATILITY,
				OPT_N*2
			);
		}
		printf("Benchmarking the reference result...\n");
	}
	printf("Binomial Option Pricinig in %s precision:\n", (doubleprecision)?"double":"single");
	printf("Pricing %d Option with time step of %d.\n", 2*OPT_N, TIMESTEP_N);
	kmp_set_defaults("KMP_AFFINITY=scatter,granularity=thread");
#pragma offload target(mic) 
		{

		}
	if (doubleprecision)	
	{
		double *CallResultList    = CallResult.DPData;
		double *StockPriceList    = StockPrice.DPData;
		double *OptionStrikeList  = OptionStrike.DPData;
		double *OptionYearsList   = OptionYears.DPData;
		printf("Benchmarking the reference result in double...\n");
		sTime = second();
#pragma offload target (mic:0) in(StockPriceList,OptionStrikeList,OptionYearsList : length(OPT_N))\
                out(CallResultList : length(OPT_N)) signal(&signal_var)
		BinomialTemplate<double>( CallResultList, StockPriceList, OptionStrikeList, OptionYearsList);
		BinomialTemplate<double>( CallResultList + OPT_N, StockPriceList + OPT_N, OptionStrikeList + OPT_N, OptionYearsList + OPT_N);
		#pragma offload_wait target(mic:0) wait(&signal_var)
		eTime = second();
		
	}
	else
	{
		float *CallResultList   = CallResult.SPData;
		float *StockPriceList   = StockPrice.SPData;
		float *OptionStrikeList = OptionStrike.SPData;
		float *OptionYearsList  = OptionYears.SPData;
		printf("Benchmarking the reference result in single...\n");
		sTime = second();
#pragma offload target (mic:0) in(StockPriceList,OptionStrikeList,OptionYearsList : length(OPT_N)) \
		out(CallResultList: length(OPT_N)) signal(&signal_var)
		BinomialTemplate<float>( CallResultList, StockPriceList, OptionStrikeList, OptionYearsList);
	        BinomialTemplate<float>( CallResultList + OPT_N, StockPriceList + OPT_N, OptionStrikeList + OPT_N, OptionYearsList + OPT_N);
		#pragma offload_wait target(mic:0) wait(&signal_var)
		eTime = second();
	}
	printf("Completed in %10.5f seconds. Options per second: %10.5f\n",eTime-sTime,  static_cast<double>(OPT_N*2)/(eTime-sTime) );


	if (verbose)
	{
		sum_delta = 0;
		sum_ref   = 0;
		max_delta = 0.0f;
		sumReserve = 0;
		int max_index = 0;

		for(i = 0; i < 2*OPT_N; i++){
			ref   = CallResultMaster[i];
			if (doubleprecision)	
				delta = fabs(CallResultMaster[i] - CallResult.DPData[i]);
			else
				delta = fabs(CallResultMaster[i] - CallResult.SPData[i]);
			if(delta > max_delta)
			{
				max_delta = delta; 
				max_index = i;
			}
			sum_delta += delta;
			sum_ref   += fabs(ref);
		}
		sumReserve /= (double)2.0*OPT_N;
		L1norm = sum_delta / sum_ref;
		if(sum_ref >1E-5)
			printf("L1 norm: %E\n", errorVal = sum_delta / sum_ref);
		else
			printf("Avg. diff: %E\n", errorVal = sum_delta / (double)(OPT_N*2));
		printf((errorVal < 5e-4) ? "TEST PASSED\n" : "TEST FAILED\n");
		scalable_aligned_free(CallResultMaster);
	}
	if (doubleprecision)	
	{
		scalable_aligned_free(CallResult.DPData);
		scalable_aligned_free(StockPrice.DPData);
		scalable_aligned_free(OptionStrike.DPData);
		scalable_aligned_free(OptionYears.DPData);
	}
	else
	{
		scalable_aligned_free(CallResult.SPData);
		scalable_aligned_free(StockPrice.SPData);
		scalable_aligned_free(OptionStrike.SPData);
		scalable_aligned_free(OptionYears.SPData);
	}
 	return 0;
}
///////////////////////////////////////////////////////////////////////////////
// Polynomial approximation of cumulative normal distribution function
///////////////////////////////////////////////////////////////////////////////
double CND(double d){
    const double       A1 = 0.31938153;
    const double       A2 = -0.356563782;
    const double       A3 = 1.781477937;
    const double       A4 = -1.821255978;
    const double       A5 = 1.330274429;
    const double RSQRT2PI = 0.39894228040143267793994605993438;

    double
        K = 1.0 / (1.0 + 0.2316419 * fabs(d));

    double
        cnd = RSQRT2PI * exp(- 0.5 * d * d) *
        (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5)))));

    if(d > 0)
        cnd = 1.0 - cnd;

    return cnd;
}

void BlackScholesBody(
    double& callResult,
    double Sf, //Stock price
    double Xf, //Option strike
    double Tf, //Option years
    double Rf, //Riskless rate
    double Vf  //Volatility rate
){
    double S = Sf, X = Xf, T = Tf, R = Rf, V = Vf;

    double sqrtT = sqrt(T);
    double    d1 = (log(S / X) + (R + 0.5 * V * V) * T) / (V * sqrtT);
    double    d2 = d1 - V * sqrtT;
    double CNDD1 = CND(d1);
    double CNDD2 = CND(d2);

    //Calculate Call and Put simultaneously
    double expRT = exp(- R * T);
    callResult   = (S * CNDD1 - X * expRT * CNDD2);
}
